#include <gtk/gtk.h>
#include <glib.h>

#include <string.h>

#include "common.h"
#include "configfile.h"
#include "browse.h"
#include "mainwin.h"
#include "query.h"
#include "util.h"

extern GtkWidget *serverlist_combo;
extern GtkWidget *statusbar;
extern struct gq_config config;



GtkWidget *create_mainwin_browsemode(void)
{
     GtkWidget *browsemode_vbox, *hbox1;
     GtkWidget *servcombo, *mainpane;
     GtkWidget *tree;
     GtkWidget *label;

     browsemode_vbox = gtk_vbox_new(FALSE, 0);
     hbox1 = gtk_hbox_new(FALSE, 0);
     gtk_widget_show(hbox1);
     gtk_box_pack_start(GTK_BOX(browsemode_vbox), hbox1, FALSE, FALSE, 3);

     /* LDAP server combo box */
     servcombo = gtk_combo_new();
     serverlist_combo = servcombo;
     fill_serverlist_combo();
     gtk_box_pack_start(GTK_BOX(hbox1), servcombo, FALSE, TRUE, 0);
     gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(servcombo)->entry), FALSE);
     gtk_signal_connect(GTK_OBJECT(GTK_COMBO(servcombo)->entry), "changed",
			GTK_SIGNAL_FUNC(servername_changed_callback),
			GTK_ENTRY(GTK_COMBO(servcombo)->entry));
     GTK_WIDGET_UNSET_FLAGS(GTK_ENTRY(GTK_COMBO(servcombo)->entry),
			    GTK_CAN_FOCUS);
     /* gtkcombo.h says don't touch this but... */
     GTK_WIDGET_UNSET_FLAGS(GTK_BUTTON(GTK_COMBO(servcombo)->button),
			    GTK_CAN_FOCUS);
     gtk_widget_show(servcombo);

     mainpane = gtk_hpaned_new();
     gtk_container_border_width(GTK_CONTAINER(mainpane), 2);
     gtk_widget_show(mainpane);
     gtk_box_pack_start(GTK_BOX(browsemode_vbox), mainpane, TRUE, TRUE, 0);

     tree = build_tree();
     if(tree) {
	  gtk_widget_set_usize(tree, 300, -1);
	  gtk_paned_add1(GTK_PANED(mainpane), tree);
     }

     label = gtk_label_new("second pane");
     gtk_widget_show(label);
     gtk_paned_add2(GTK_PANED(mainpane), label);



     gtk_widget_show(browsemode_vbox);

     return(browsemode_vbox);
}


GtkWidget *build_tree(void)
{
     GtkWidget *root, *tree, *server_item, *item;
     struct ldapserver *server;
     int i;
     char suffix[MAX_NUM_SUFFIX + 1][MAX_SUFFIX_LEN];

     server = server_by_name(config.cur_servername);
     if(!server) {
          statusbar_msg("Oops! No server found!");
          return(NULL);
     }

     root = gtk_tree_new();
     gtk_widget_show(root);

     server_item = gtk_tree_item_new_with_label(server->name);
     gtk_widget_show(server_item);
     gtk_tree_append(GTK_TREE(root), server_item);

     tree = gtk_tree_new();
     gtk_tree_item_set_subtree(GTK_TREE_ITEM(server_item), tree);

     get_suffix(server, suffix);

     for(i = 0; suffix[i][0]; i++) {
	  printf("got: |%s|\n", suffix[i]);

	  item = gtk_tree_item_new_with_label(suffix[i]);
	  gtk_widget_show(item);
	  gtk_tree_append(GTK_TREE(tree), item);

     }

     /* open up tree for current host */
     gtk_tree_item_expand(GTK_TREE_ITEM(server_item));

     return(root);
}


/*
 * get all suffixes a server considers itself authorative for.
 */
void get_suffix(struct ldapserver *server, char suffix[MAX_NUM_SUFFIX + 1][MAX_SUFFIX_LEN])
{
     LDAP *ld;
     LDAPMessage *res, *e;
     BerElement *ptr;
     int msg, i, j, s;
     char *attr, **vals;
     char *ldapv3_config[] = {
	  "namingcontexts",
	  NULL
     };

     suffix[0][0] = 0;
     ld = open_connection(server);
     if(ld == NULL)
	  return;

     /* try Umich/OpenLDAP style config */
     msg = ldap_search_s(ld, "cn=config", LDAP_SCOPE_BASE,
			 "(objectclass=*)", NULL, 0, &res);

     if(msg == LDAP_SUCCESS) {

	  /* only expecting one attribute type (database) with one
	     value here, no checking done on length or number of suffixes */
	  e = ldap_first_entry(ld, res);
	  if(e) {
	       attr = ldap_first_attribute(ld, e, &ptr);
	       if(!strcasecmp(attr, "database")) {
		    vals = ldap_get_values(ld, e, attr);
		    i = s = 0;
		    /* skip first value (backend type) */
		    while(vals[0][i] && vals[0][i] != ':')
			 i++;
		    if(vals[0][i] == ':')
			 i++;
		    while(vals[0][i]) {
			 j = 0;
			 while(vals[0][i] && vals[0][i] != ':')
			      /* skip whitespace prefix */
			      if(j || vals[0][i] != ' ')
				   suffix[s][j++] = vals[0][i++];
			      else
				   i++;
			 /* delete trailing spaces */
			 while(suffix[s][j - 1] == ' ')
			      j--;
			 suffix[s][j] = 0;
			 s++;
			 i++;
		    }
		    suffix[s][0] = 0;
	       }
	  }
     }

     if(!suffix[0][0]) {

	  /* try LDAP V3/Netscape style config */
	  msg = ldap_search_s(ld, "", LDAP_SCOPE_BASE, "(objectclass=*)",
			      ldapv3_config, 0, &res);
	  if(msg != LDAP_SUCCESS)
	       return;

	  e = ldap_first_entry(ld, res);
	  if(!e)
	       return;

	  for(attr = ldap_first_attribute(ld, e, &ptr); attr;
	      attr = ldap_next_attribute(ld, e, ptr)) {
	       if(!strcasecmp(attr, "namingcontexts")) {
		    fflush(stdout);
		    vals = ldap_get_values(ld, e, attr);
		    for(i = 0; vals[i] && i < MAX_NUM_SUFFIX; i++) {
			 strncpy(suffix[i], vals[i], MAX_SUFFIX_LEN);
		    }
		    suffix[i][0] = 0;
	       }
	  }

     }

}

